import io
import os
import shutil
import tempfile
from typing import Optional
from uuid import UUID, uuid4
import hashlib
import celery
import requests
from billiard.exceptions import SoftTimeLimitExceeded
from django.db.transaction import atomic

from video_transcoding import models, transcoding, defaults
from video_transcoding.celery import app
from video_transcoding.utils import LoggerMixin

DESTINATION_FILENAME = '{basename}1080p.mp4'

CONNECT_TIMEOUT = 1
DOWNLOAD_TIMEOUT = 60 * 60
UPLOAD_TIMEOUT = 60 * 60


class TranscodeVideo(LoggerMixin, celery.Task):
    """ Video processing task."""
    name = 'video.transcode'
    routing_key = 'video_transcoding'

    def run(self, video_id: int, download: bool = False) -> Optional[str]:
        """
        Process video.

        1. Locks video changing status from QUEUED to PROCESS
        2. Transcodes video and stores result to origins
        3. Changes video status to DONE, stores result basename
        4. On errors changes video status ERROR, stores error message

        :param video_id: Video id.
        :param download: Download source file to tmp dir before processing.
        """
        status = models.Video.DONE
        error = basename = None
        video = self.lock_video(video_id)
        try:
            basename = uuid4().hex
            self.process_video(video, basename, download=download)
        except SoftTimeLimitExceeded as e:
            # celery graceful shutdown
            status = models.Video.QUEUED
            basename = None
            error = repr(e)
            raise self.retry(countdown=10)
        except Exception as e:
            basename = None
            status = models.Video.ERROR
            error = repr(e)
        finally:
            self.unlock_video(video_id, status, error, basename)
        return error

    def select_for_update(self, video_id: int, status: int) -> models.Video:
        """ Lock video in DB for current task.

        :param video_id: Video primary key
        :param status: expected video status
        :returns: Video object from db

        :raises models.Video.DoesNotExist: in case of missing or locked
            Video for primary key
        :raises ValueError: in case of unexpected Video status or task_id

        """
        try:
            video = models.Video.objects.select_for_update(
                skip_locked=True, of=('self',)).get(pk=video_id)
        except models.Video.DoesNotExist:
            self.logger.error("Can't lock video %s", video_id)
            raise

        if video.task_id != UUID(self.request.id):
            self.logger.error("Unexpected video %s task_id %s",
                              video.id, video.task_id)
            raise ValueError(video.task_id)

        if video.status != status:
            self.logger.error("Unexpected video %s status %s",
                              video.id, video.get_status_display())
            raise ValueError(video.status)
        return video

    @atomic
    def lock_video(self, video_id: int) -> models.Video:
        """
        Gets video in QUEUED status from DB and changes status to PROCESS.

        :param video_id: Video primary key
        :returns: Video object
        :raises Retry: in case of unexpected video status or task_id
        """
        try:
            video = self.select_for_update(video_id, models.Video.QUEUED)
        except (models.Video.DoesNotExist, ValueError) as e:
            # if video is locked or task_id is not equal to current task, retry.
            raise self.retry(exc=e)

        video.change_status(models.Video.PROCESS)
        return video

    @atomic
    def unlock_video(self, video_id: int, status: int, error: Optional[str],
                     basename: Optional[str]) -> None:
        """
        Marks video with final status.

        :param video_id: Video primary key
        :param status: final video status (Video.DONE, Video.ERROR)
        :param error: error message
        :param basename: UUID-like result file identifier
        :raises RuntimeError: in case of unexpected video status or task id
        """
        try:
            video = self.select_for_update(video_id, models.Video.PROCESS)
        except (models.Video.DoesNotExist, ValueError) as e:
            # if video is locked or task_id differs from current task, do
            # nothing because video is modified somewhere else.
            raise RuntimeError("Can't unlock locked video %s: %s",
                               video_id, repr(e))

        video.change_status(status, error=error, basename=basename)

    def process_video(self, video: models.Video, basename: str,
                      download: bool = False) -> None:
        """
        Video processing workflow.

        1. Create temporary directory
        2. Transcode source file
        3. Upload resulting file to origins
        4. Cleanup temporary directory

        :param video: Video object
        :param basename: video files common base name
        :param download: download source to temp dir
        """
        with tempfile.TemporaryDirectory(dir=defaults.VIDEO_TEMP_DIR,
                                         prefix=f'video-{video.pk}-') as d:
            destination = os.path.join(d, f'{basename}1080p.mp4')
            if download:
                source = os.path.join(d, f'{basename}.src.bin')
                self.download(video.source, source)
            else:
                source = video.source
            self.transcode(source, destination)
            self.store(destination)
        self.logger.info("Processing done")

    def download(self, source: str, destination: str) -> None:
        """
        Downloads source to temporary directory
        :param source: source file link
        :param destination: path to downloaded file
        """
        self.logger.info("Start downloading %s to %s", source, destination)
        timeout = (CONNECT_TIMEOUT, DOWNLOAD_TIMEOUT)
        if defaults.CHECKSUM_SOURCE:
            # noinspection PyUnresolvedReferences,PyProtectedMember
            checksum = hashlib.md5()  # type: Optional[hashlib._Hash]
        else:
            checksum = None
        with requests.get(source, stream=True, timeout=timeout, allow_redirects=True) as response:
            response.raise_for_status()
            with open(destination, 'wb') as f:
                encoding = response.headers.get('transfer-encoding')
                if encoding or checksum:
                    self.logger.warning(
                        "Transfer-encoding is %s, not fastest one",
                        encoding or checksum)
                    for chunk in response.iter_content(io.DEFAULT_BUFFER_SIZE):
                        f.write(chunk)
                        if checksum:
                            checksum.update(chunk)
                else:
                    shutil.copyfileobj(response.raw, f)
                content_length = response.headers.get('Content-Length')
                if content_length is not None:
                    size = f.tell()
                    if size != int(content_length):
                        raise ValueError("Partial file", size)
        self.logger.info("Downloading %s finished", source)
        if checksum:
            self.logger.info("Source file checksum: %s", checksum.hexdigest())

    def transcode(self, source: str, destination: str) -> None:
        """
        Starts video transcoding

        :param source: source file link (http/ftp or file path)
        :param destination: result temporary file path.
        """
        self.logger.info("Start transcoding %s to %s",
                         source, destination)
        transcoder = transcoding.Transcoder(source, destination)
        transcoder.transcode()
        self.logger.info("Transcoding %s finished", source)

    def store(self, destination: str) -> None:
        """
        Stores transcoded video to origin list

        :param destination: transcoded video path.
        """
        self.logger.info("Start saving %s to origins", destination)
        filename = os.path.basename(destination)
        timeout = (CONNECT_TIMEOUT, UPLOAD_TIMEOUT)
        for origin in defaults.VIDEO_ORIGINS:
            url = os.path.join(origin, filename)
            self.logger.debug("Uploading %s to %s", destination, url)
            with open(destination, 'rb') as f:
                response = requests.put(url, data=f, timeout=timeout)
                response.raise_for_status()
            self.logger.info("Uploaded to %s", url)
        self.logger.info("%s save finished", destination)


transcode_video: TranscodeVideo = app.register_task(
    TranscodeVideo())  # type: ignore



import os

from celery.result import AsyncResult

from video_transcoding import tasks, defaults
from video_transcoding import models


def send_transcode_task(video: models.Video) -> AsyncResult:
    """
    Send a video transcoding task.

    If task is successfully sent to broker, Video status is changed to QUEUED
    and Celery task identifier is saved.

    :param video: video object
    :type video: video.models.Video
    :returns: Celery task result
    :rtype: celery.result.AsyncResult
    """
    result = tasks.transcode_video.apply_async(
        args=(video.pk,),
        kwargs={'download': defaults.VIDEO_DOWNLOAD_SOURCE},
        countdown=10)
    video.change_status(video.QUEUED, task_id=result.task_id)
    return result



import os
from typing import Any, cast
from uuid import UUID

from django.core.validators import URLValidator
from django.db import models
from django.utils.translation import gettext_lazy as _
from model_utils.models import TimeStampedModel

from video_transcoding import defaults


class Video(TimeStampedModel):
    """ Video model."""
    CREATED, QUEUED, PROCESS, DONE, ERROR = range(5)
    STATUS_CHOICES = (
        (CREATED, _('created')),  # And editor created video in db
        (QUEUED, _('queued')),  # Transcoding task is sent to broker
        (PROCESS, _('process')),  # Celery worker started video processing
        (DONE, _('done')),  # Video processing is done successfully
        (ERROR, _('error')),  # Video processing error
    )

    status = models.SmallIntegerField(default=CREATED, choices=STATUS_CHOICES,
                                      verbose_name=_('Status'))
    error = models.TextField(blank=True, null=True, verbose_name=_('Error'))
    task_id = models.UUIDField(blank=True, null=True,
                               verbose_name=_('Task ID'))
    source = models.URLField(verbose_name=_('Source'),
                             validators=[
                                 URLValidator(schemes=('http', 'https'))])
    basename = models.UUIDField(blank=True, null=True,
                                verbose_name=_('Basename'))

    class Meta:
        app_label = 'video_transcoding'
        verbose_name = _('Video')
        verbose_name_plural = _('Video')

    def __str__(self) -> str:
        basename = os.path.basename(self.source)
        return f'{basename} ({self.get_status_display()})'

    def format_video_url(self, edge: str) -> str:
        """
        Returns a link to m3u8 playlist on one of randomly chosen edges.
        """
        if self.basename is None:
            raise RuntimeError("Video has no files")
        basename = cast(UUID, self.basename)
        return defaults.VIDEO_URL.format(
            edge=edge.rstrip('/'),
            filename=basename.hex)

    def change_status(self, status: int, **fields: Any) -> None:
        """
        Changes video status.

        Also saves another model fields and always updates `modified` value.

        :param status: one of statuses for Video.status (see STATUS_CHOICES)
        :param fields: dict with model field values.
        """
        self.status = status
        update_fields = {'status', 'modified'}
        for k, v in fields.items():
            setattr(self, k, v)
            update_fields.add(k)
        self.save(update_fields=tuple(update_fields))



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class VideoTranscodingConfig(AppConfig):
    name = 'video_transcoding'
    verbose_name = _('Video Transcoding')

    def ready(self) -> None:
        __import__('video_transcoding.signals')



from logging import getLogger
from typing import Any


class LoggerMixin:
    """
    A mixin for logger injection.

    Should not be used with Django models, because Logger contains
    non-serializable threading.Lock object.
    """
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        cls = self.__class__
        self.logger = getLogger(f'{cls.__module__}.{cls.__name__}')



import random
from typing import Any, TypeVar, Callable

from django.contrib import admin
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from video_transcoding import helpers, models, defaults

C = TypeVar("C", bound=Callable)


def short_description(name: str) -> Callable[[C], C]:
    """ Sets short description for function."""

    def inner(func: C) -> C:
        setattr(func, 'short_description', name)
        return func

    return inner


@admin.register(models.Video)
class VideoAdmin(admin.ModelAdmin):
    list_display = ('basename', 'source', 'status_display')
    list_filter = ('status',)
    search_fields = ('source', '=basename')
    actions = list(admin.ModelAdmin.actions or ()) + ["transcode"]
    readonly_fields = ('created', 'modified', 'video_player')

    class Media:
        _prefix = 'https://cdn.jsdelivr.net/mediaelement/latest'
        css = {
            "all": (f'{_prefix}/mediaelementplayer.css',)
        }
        js = (f'{_prefix}/mediaelement-and-player.min.js',)

    @short_description(_("Status"))
    def status_display(self, obj: models.Video) -> str:
        return obj.get_status_display()

    # noinspection PyUnusedLocal
    @short_description(_('Send transcode task'))
    def transcode(self,
                  request: HttpRequest,
                  queryset: "QuerySet[models.Video]") -> None:
        for video in queryset:
            helpers.send_transcode_task(video)

    @short_description(_('Video player'))
    def video_player(self, obj: models.Video) -> str:
        if obj.basename is None:
            return ""
        edge = random.choice(defaults.VIDEO_EDGES)
        source = obj.format_video_url(edge)
        return mark_safe('''
<video width="480px" height="270px" class="mejs__player">
<source src="%s" />
</video>''' % (source,))

    def add_view(self,
                 request: HttpRequest,
                 form_url: str = '',
                 extra_context: Any = None
                 ) -> HttpResponse:
        fields, self.fields = self.fields, ('source',)
        try:
            return super().add_view(request, form_url, extra_context)
        finally:
            self.fields = fields



import os
import signal
from typing import Any

from celery import Celery
from celery.signals import worker_shutting_down
from celery.utils.log import get_logger
from django.conf import settings

from video_transcoding import defaults

app = Celery(defaults.CELERY_APP_NAME)
app.config_from_object(defaults.VIDEO_TRANSCODING_CELERY_CONF)
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)


# noinspection PyUnusedLocal
@worker_shutting_down.connect
def send_term_to_children(**kwargs: Any) -> None:
    get_logger(app.__module__).warning(
        "Received shutdown signal, sending SIGUSR1 to worker process group")
    # raises SoftTimeLimitExceeded in worker processes
    os.killpg(os.getpid(), signal.SIGUSR1)



from os import getenv as e

from kombu import Queue


CELERY_APP_NAME = 'video_transcoding'


VIDEO_TRANSCODING_CELERY_CONF = {
    'broker_url': e('VIDEO_TRANSCODING_CELERY_BROKER_URL',
                    'amqp://guest:guest@rabbitmq:5672/'),
    'result_backend': e('VIDEO_TRANSCODING_CELERY_RESULT_BACKEND', None),
    'task_default_exchange': CELERY_APP_NAME,
    'task_default_exchange_type': 'topic',
    'task_default_queue': CELERY_APP_NAME,
    'worker_prefetch_multiplier': 1,
    'worker_concurrency': e('VIDEO_TRANSCODING_CELERY_CONCURRENCY'),
    'task_acks_late': True,
    'task_reject_on_worker_lost': True,
    'task_queues': [
        Queue(CELERY_APP_NAME, routing_key=CELERY_APP_NAME),
    ]
}

# Directory for large output files
VIDEO_TEMP_DIR = '/tmp'

# Download source before processing
VIDEO_DOWNLOAD_SOURCE = bool(int(e('VIDEO_DOWNLOAD_SOURCE', 0)))

# A list of WebDAV endpoints for storing video results
VIDEO_ORIGINS = e('VIDEO_ORIGINS',
                  'http://storage.localhost:8080/videos/').split(',')

# Video streamer public urls (comma-separated)
VIDEO_EDGES = e('VIDEO_EDGES', 'http://storage.localhost:8080/').split(',')

# Edge video manifest url template
VIDEO_URL = '{edge}/hls/{filename}1080p.mp4/index.m3u8'

# Output source files checksum
CHECKSUM_SOURCE = bool(int(e('CHECKSUM_SOURCE', 0)))



import math
from dataclasses import dataclass
from pprint import pformat
from typing import Optional, Dict, Any, Tuple, cast

import pymediainfo
from fffw.encoding import codecs
from fffw.graph.meta import VideoMeta, AudioMeta, AUDIO, VIDEO, from_media_info, TS
from fffw.encoding.inputs import input_file, Stream
from fffw.encoding.filters import Scale
from fffw.encoding.outputs import output_file
from fffw.encoding.ffmpeg import FFMPEG
from fffw.wrapper import param

from video_transcoding.utils import LoggerMixin

AUDIO_CODEC = 'audio_codec'
VIDEO_CODEC = 'video_codec'
SCALE = 'scale'

# Used metadata keys
AUDIO_DURATION = 'audio_duration'
AUDIO_SAMPLING_RATE = 'audio_sampling_rate'
SAMPLES_COUNT = 'samples'
VIDEO_DURATION = 'video_duration'
VIDEO_FRAME_RATE = 'video_frame_rate'
FRAMES_COUNT = 'frames'
DAR = 'aspect'
PAR = 'par'
WIDTH = 'width'
HEIGHT = 'height'

# Handy media info log format
MEDIA_INFO_MSG_FORMAT = """%s media info:
VIDEO:
%s
AUDIO: 
%s
"""

# HLS Segment duration step, seconds
SEGMENT_SIZE = 4
# H.264 Group of pixels duration, seconds
GOP_DURATION = 2

# Force key frame every N seconds
KEY_FRAMES = 'expr:if(isnan(prev_forced_t),1,gte(t,prev_forced_t+{sec}))'

# Allowed duration difference between source and result
DURATION_DELTA = 0.95

# Video transcoding params
TRANSCODING_OPTIONS = {
    VIDEO_CODEC: {
        'codec': 'libx264',
        'force_key_frames': KEY_FRAMES.format(sec=SEGMENT_SIZE),
        'constant_rate_factor': 23,
        'preset': 'slow',
        'max_rate': 5_000_000,
        'buf_size': 10_000_000,
        'profile': 'high',
        'pix_fmt': 'yuv420p',
    },
    SCALE: {
        'width': 1920,
        'height': 1080,
    },
    AUDIO_CODEC: {
        'codec': 'aac',
        'bitrate': 192000,
        'channels': 2,
    },
}

Metadata = Dict[str, Any]
""" File metadata type."""


@dataclass
class AudioCodec(codecs.AudioCodec):
    rate: float = param(name='ar')
    channels: int = param(name='ac')


@dataclass
class VideoCodec(codecs.VideoCodec):
    force_key_frames: str = param()
    constant_rate_factor: int = param(name='crf')
    preset: str = param()
    max_rate: int = param(name='maxrate')
    buf_size: int = param(name='bufsize')
    profile: str = param(stream_suffix=True)
    gop: int = param(name='g')
    rate: float = param(name='r')
    pix_fmt: str = param()


class TranscodeError(Exception):
    """ Video transcoding error."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class Transcoder(LoggerMixin):
    """ Video transcoder.

    >>> t = Transcoder('http://source.localhost/source.mp4', '/tmp/result.mp4')
    >>> t.transcode()
    """

    def __init__(self, source: str, destination: str):
        """
        :param source: source file link (http/ftp or file path)
        :param destination: result file path
        """
        super().__init__()
        self.source = source
        self.destination = destination

    def get_media_info(self, video: Optional[VideoMeta],
                       audio: Optional[AudioMeta]) -> Metadata:
        """
        Transforms video and audio metadata to a dict

        :param video: video stream metadata
        :param audio: audio stream metadata
        :returns: metadata single level dictionary
        """
        if video is None:
            raise TranscodeError("missing video stream")
        if audio is None:
            raise TranscodeError("missing audio stream")

        media_info = {
            'width': video.width,
            'height': video.height,
            'aspect': video.dar,
            'par': video.par,
            VIDEO_DURATION: video.duration.total_seconds(),
            'video_bitrate': video.bitrate,
            VIDEO_FRAME_RATE: video.frame_rate,
            'audio_bitrate': audio.bitrate,
            AUDIO_SAMPLING_RATE: audio.sampling_rate,
            AUDIO_DURATION: audio.duration.total_seconds(),
        }
        self.logger.info("Parsed media info:\n%s", pformat(media_info))
        return media_info

    def transcode(self) -> None:
        """ Transcodes video

        * checks source mediainfo
        * runs `ffmpeg`
        * validates result
        """
        audio_meta, video_meta = self.get_meta_data(self.source)

        # Get source mediainfo to use in validation
        source_media_info = self.get_media_info(video_meta, audio_meta)

        # set group of pixels length to segment size
        gop = math.floor(source_media_info[VIDEO_FRAME_RATE] * GOP_DURATION)
        # preserve original video FPS
        vrate = source_media_info[VIDEO_FRAME_RATE]
        # preserve source audio sampling rate
        arate = source_media_info[AUDIO_SAMPLING_RATE]

        # Common ffmpeg flags
        ff = FFMPEG(overwrite=True, loglevel='repeat+level+info')
        # Init source file
        ff < input_file(self.source,
                        Stream(VIDEO, video_meta),
                        Stream(AUDIO, audio_meta))

        # Output codecs
        video_opts = cast(Dict[str, Any], TRANSCODING_OPTIONS[VIDEO_CODEC])
        cv0 = VideoCodec(
            gop=gop,
            rate=vrate,
            **video_opts)
        audio_opts = cast(Dict[str, Any], TRANSCODING_OPTIONS[AUDIO_CODEC])
        ca0 = AudioCodec(
            rate=arate,
            **audio_opts)

        # Scaling
        ff.video | Scale(**TRANSCODING_OPTIONS[SCALE]) > cv0

        # codecs, muxer and output path
        ff > output_file(self.destination, cv0, ca0, format='mp4')

        # Run ffmpeg
        self.run(ff)

        # Get result mediainfo
        audio_meta, video_meta = self.get_meta_data(self.destination)
        dest_media_info = self.get_media_info(video_meta, audio_meta)

        # Validate ffmpeg result
        self.validate(source_media_info, dest_media_info)

    def fix_par(self, t: pymediainfo.Track) -> None:
        """
        Fix PAR to satisfy equation DAR = width / height * PAR.
        """
        dar = float(t.__dict__.get('display_aspect_ratio', 0))
        par = float(t.__dict__.get('pixel_aspect_ratio', 0))
        width = int(t.__dict__.get('width', 0))
        height = int(t.__dict__.get('height', 0))

        if not (width and height):
            # not enough info to fix PAR
            self.logger.debug("width or height unknown, can't restore PAR metadata")
            return
        ratio = width / height

        if not par and dar:
            self.logger.debug("restoring par from dar")
            par = dar / ratio
        elif not dar and par:
            self.logger.debug("restoring dar from par")
            dar = par * ratio
        elif not dar and not par:
            self.logger.debug("setting aspects to defaults")
            par = 1.0
            dar = ratio

        # at this moment we know all 4 variables, checking equation
        if abs(dar - ratio * par) >= 0.001:  # see fffw.meta.VideoMeta.validate
            # par is least reliable value, using it to fix equation
            par = dar / ratio
        t.__dict__['display_aspect_ratio'] = f'{dar:.3f}'
        t.__dict__['pixel_aspect_ratio'] = f'{par:.3f}'

    def fix_frames(self, t: pymediainfo.Track) -> None:
        """
        Fix frames count to satisfy equation:

        Duration = FPS * frames
        """
        duration = float(t.__dict__.get('duration', 0)) / 1000  # duration in seconds
        frame_rate = float(t.__dict__.get('frame_rate', 0))
        frames = int(t.__dict__.get('frame_count', 0))

        if not duration and frames and frame_rate:
            self.logger.debug("restoring video duration")
            duration = frames / frame_rate
        elif not frames and duration and frame_rate:
            self.logger.debug("restoring frames")
            frames = round(duration * frame_rate)
        elif not frame_rate and duration and frames:
            self.logger.debug("restoging frame_rate")
            frame_rate = frames / duration
        elif not all([frames, frame_rate, duration]):
            # 2 of 3 variables are unknown, or even all of them.
            # can't restore metadata
            return

        # checking equation
        if abs(frames - duration * frame_rate) > 1:
            # frames is least reliable value
            frame_rate = frames / duration

        t.__dict__['frame_rate'] = f'{frame_rate:.3f}'
        t.__dict__['duration'] = f'{duration * 1000.0:.3f}'  # milliseconds again
        t.__dict__['frame_count'] = f'{frames}'

    def fix_samples(self, t: pymediainfo.Track) -> None:
        """
        Fix sample count to satisfy equation:

        Duration = Sampling rate * samples
        """
        duration = float(t.__dict__.get('duration', 0)) / 1000  # duration in seconds
        sampling_rate = float(t.__dict__.get('sampling_rate', 0))
        samples = int(t.__dict__.get('samples_count', 0))

        if not duration and samples and sampling_rate:
            self.logger.debug("restoring audio duration")
            duration = samples / sampling_rate
        elif not samples and duration and sampling_rate:
            self.logger.debug("restoring samples")
            samples = round(duration * sampling_rate)
        elif not sampling_rate and duration and samples:
            self.logger.debug("restoging sampling_rate")
            sampling_rate = samples / duration
        elif not all([samples, sampling_rate, duration]):
            # 2 of 3 variables are unknown, or even all of them.
            # can't restore metadata
            return

        # fix sampling rate type
        sampling_rate = round(sampling_rate)
        # handle duration rounding
        duration = round(duration, 3)

        # checking equation
        if abs(samples - duration * sampling_rate) > 1:
            # samples is least reliable data, because sampling rate has common values like 48000,
            # and duration is more reliable.
            samples = round(duration * sampling_rate)

        t.__dict__['sampling_rate'] = f'{sampling_rate}'
        t.__dict__['duration'] = f'{duration * 1000.0:.3f}'  # milliseconds again
        t.__dict__['samples_count'] = f'{samples}'

    def get_meta_data(self, filename: str
                      ) -> Tuple[Optional[AudioMeta], Optional[VideoMeta]]:
        result: pymediainfo.MediaInfo = pymediainfo.MediaInfo.parse(filename)
        for t in result.tracks:
            if t.track_type in ('Video', 'Image'):
                self.fix_par(t)
                self.fix_frames(t)
            elif t.track_type == 'Audio':
                self.fix_samples(t)
        metadata = from_media_info(result)
        video_meta = None
        audio_meta = None
        for m in metadata:
            if m.kind == VIDEO and video_meta is None:
                video_meta = m
            if m.kind == AUDIO and audio_meta is None:
                audio_meta = m
        return audio_meta, video_meta

    @staticmethod
    def validate(source_media_info: Metadata,
                 dest_media_info: Metadata) -> None:
        """
        Validate video transcoding result.

        :param source_media_info: source metadata
        :param dest_media_info: result metadata
        """
        src_duration = max(source_media_info[VIDEO_DURATION],
                           source_media_info[AUDIO_DURATION])
        dst_duration = min(dest_media_info[VIDEO_DURATION],
                           dest_media_info[AUDIO_DURATION])
        if dst_duration < DURATION_DELTA * src_duration:
            # Check whether result duration corresponds to source duration
            # (damaged source files may be processed successfully but result
            # is shorter)
            raise TranscodeError(f"incomplete file: {dst_duration}")

    @staticmethod
    def run(ff: FFMPEG) -> None:
        """ Starts ffmpeg process and captures errors from it's logs"""
        return_code, output, error = ff.run()
        if return_code != 0:
            # Check return code and error messages
            error = error or f"invalid ffmpeg return code {return_code}"
            raise TranscodeError(error)



# Generated by Django 3.0.3 on 2020-02-26 09:19

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('video_transcoding', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='video',
            name='source',
            field=models.URLField(validators=[django.core.validators.URLValidator(schemes=('http', 'https'))]),
        ),
    ]



# Generated by Django 3.0.3 on 2020-02-10 05:42

import django.core.validators
from django.db import migrations, models
import django.utils.timezone
import model_utils.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Video',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('status', models.SmallIntegerField(choices=[(0, 'created'), (1, 'queued'), (2, 'process'), (3, 'done'), (4, 'error')], default=0)),
                ('error', models.TextField(blank=True, null=True)),
                ('task_id', models.UUIDField(blank=True, null=True)),
                ('source', models.URLField(validators=[django.core.validators.URLValidator(schemes=('ftp', 'http'))])),
                ('basename', models.UUIDField(blank=True, null=True)),
            ],
            options={
                'verbose_name': 'Video',
                'verbose_name_plural': 'Video',
            },
        ),
    ]



from typing import Any

from django.db import transaction
from django.db.models.signals import post_save
from django.dispatch import receiver

from video_transcoding import helpers, models


# noinspection PyUnusedLocal
@receiver(post_save, sender=models.Video)
def send_transcode_task(sender: Any, *, instance: models.Video, created: bool,
                        **kw: Any) -> None:
    if not created:
        return
    transaction.on_commit(lambda: helpers.send_transcode_task(instance))



# Generated by Django 3.0.6 on 2020-05-25 11:30

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('video_transcoding', '0002_auto_20200226_0919'),
    ]

    operations = [
        migrations.AlterField(
            model_name='video',
            name='basename',
            field=models.UUIDField(blank=True, null=True, verbose_name='Basename'),
        ),
        migrations.AlterField(
            model_name='video',
            name='error',
            field=models.TextField(blank=True, null=True, verbose_name='Error'),
        ),
        migrations.AlterField(
            model_name='video',
            name='source',
            field=models.URLField(validators=[django.core.validators.URLValidator(schemes=('http', 'https'))], verbose_name='Source'),
        ),
        migrations.AlterField(
            model_name='video',
            name='status',
            field=models.SmallIntegerField(choices=[(0, 'created'), (1, 'queued'), (2, 'process'), (3, 'done'), (4, 'error')], default=0, verbose_name='Status'),
        ),
        migrations.AlterField(
            model_name='video',
            name='task_id',
            field=models.UUIDField(blank=True, null=True, verbose_name='Task ID'),
        ),
    ]
